//
// Anti-spam filter utility.
// Copyright (C) 2004-2007 Serge Vakulenko, <vak@cronyx.ru>
//
// Usage: mk_black [--clean]
//
// Compile these sources using GNU D compiler:
//	gdc -O mk_black.d -o mk_black
//
// This file is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.
//
// You can redistribute this file and/or modify it under the terms of the GNU
// General Public License (GPL) as published by the Free Software Foundation;
// either version 2 of the License, or (at your discretion) any later version.
// See the accompanying file "COPYING.txt" for more details.
//
import std.stdio;
import std.stream;
import std.string;
import std.file;
import std.regexp : RegExp;
import etc.c.zlib;

int print_boxname_flag;
int remove_unused_flag;
char[][] blackhosts;
char[][] blackemails;
RegExp[] blackhost_patterns;
char[][] loghosts;
char[][] logemails;
int blackhosts_cnt;
int blackemails_cnt;
int blackhost_patterns_cnt;
int loghosts_cnt;
int logemails_cnt;

const char[] HOST_BLACK_LIST = "/etc/mail/hosts-black";
const char[] EMAIL_BLACK_LIST = "/etc/mail/emails-black";

const char[][] MYRELAYS = [
	"hanoi.cronyx.ru",
	"kucha.rinet.ru",
	"dragon.relcom.ru",
	"relay1.relcom.ru",
	"relay.cronyx.ru",
	"gianda.cronyx.ru",
	"relay.rinet.ru",
	"imfiko.bishkek.su",
	"mtu-net.ru",
	"corbina.ru",
	"hueymiccailhuitl.mtu.ru",
	"umail.ru",
	"hut.comstar.ru",
	"ns.sovintel.ru",
	"relay.mplik.ru"];

const char[][] WHITEHOSTS = [
	r"\.cronyx\.ru$",
	r"\.rinet\.ru$",
	r"\.mail\.ru$",
	r"\.yandex\.ru$",
	r"\.rambler\.ru$",
	r"\.tochka\.ru$",
	r"\.yahoo\.com$",
	r"\.gmail\.com$",
	r"\.gmail\.ru$",
	r"\.google\.com$",
	r"barbuda\.mtu\.ru$",
	r"umail\.mtu\.ru$",
	r"chen\.mtu\.ru$",
	r"lists\.sourceforge\.net$",
	r"externalmx-.*\.sourceforge\.net$",
	r"\.mx\.aol\.com$",
	r"smtp\.rol\.ru$",
	r"\.holmes\.peterlink\.ru$"];

const char[][] WHITEMAILS = [
	"gmail.ru",
	"mail.ru",
	"narod.ru",
	"yandex.ru",
	"rambler.ru",
	"pochta.ru",
	"mtu-net.ru",
	"aport.ru",
	"gmail.com",
	"nokia.com",
	"adobe.com",
	"paypal.com",
	"hotmail.com",
	"aol.com",
	"sun.com",
	"lycos.com",
	"microsoft.com",
	"usa.net",
	"netscape.net",
	"hinet.net",
	"yahoo.co.uk"
];

void add_host_pattern (char[] address)
{
	int n = address.find ('#');
	if (n >= 0)
		address = address [0..n];

	address = address.strip();
	if (address.length <= 0)
		return;

	address = address.replace (".", "\\.");
	address = address.replace ("?", ".");
	if (blackhost_patterns.length <= blackhost_patterns_cnt)
		blackhost_patterns.length = blackhost_patterns.length + 100;
	blackhost_patterns [blackhost_patterns_cnt++] =
		new RegExp (address ~ "$");
}

int find_host_pattern (char[] address)
{
	for (int n=0; n<blackhost_patterns_cnt; ++n) {
		if (blackhost_patterns[n].find (address) >= 0)
			return 1;
	}
	return 0;
}

void scan_for_host (char[] line, char[] boxname)
{
	if (RegExp ("^Received: *from").find (line) < 0)
		return;

	if (RegExp (r"localhost *\( *localhost *\)").find (line) >= 0)
		return;

	// Skip header, when passed not via known server.
	static RegExp regexp_myrelays;
	if (! regexp_myrelays)
		regexp_myrelays = new RegExp ("by *(" ~
			MYRELAYS.join("|") ~ ")");
	if (regexp_myrelays.find (line) < 0)
		return;

	if (RegExp (r"\(localhost \[127\.0\.0\.1]\)").find (line) >= 0)
		return;

	char[] address;
	address = RegExp (r"[^(]*\(").replace (line, "");
	address = RegExp ("([^[ ]*).*").replace (address, "$1");
	address = address.strip().tolower();
	foreach (char[] r; MYRELAYS)
		if (address.cmp (r) == 0)
			return;

	// Check that address is like x.y
	if (RegExp (r".\..").find (address) < 0)
		return;

	foreach (char[] h; WHITEHOSTS)
		if (RegExp (h).find (address) >= 0)
			return;

	if (find_host_pattern (address))
		return;

	// Add new host name to a list.
	if (blackhosts.length <= blackhosts_cnt)
		blackhosts.length = blackhosts.length + 100;
	blackhosts [blackhosts_cnt++] = address;
	append_loghosts (address);

	add_host_pattern (address);
	if (print_boxname_flag) {
		writefln ("\nBy %s:", boxname);
		print_boxname_flag = 0;
	}
	writefln ("New host: %s", address);
}

void scan_for_email (char[] line, char[] boxname)
{
	if (RegExp (r"\(envelope-from.*\)").find (line) < 0)
		return;

	char[] address;
	address = RegExp (r".*\(envelope-from(.*)\).*").replace (line, "$1");
	address = RegExp ("<(.*)>").replace (address.strip(), "$1");

	// Check that address is x@y
	if (address.find ('@') < 0)
		return;
	address = address.tolower();

	if (RegExp (r"[.@]cronyx\.ru").find (address) >= 0)
		return;
	if (RegExp (r"\.kiae\.su").find (address) >= 0)
		return;
	if (address.find (' ') >= 0)
		return;

	for (int n=0; n<blackemails_cnt; ++n)
		if (address.cmp (blackemails [n]) == 0)
			return;

	// Add new e-mail address to a list.
	if (blackemails.length <= blackemails_cnt)
		blackemails.length = blackemails.length + 100;
	blackemails [blackemails_cnt++] = address;
	append_logemails (address);

	if (print_boxname_flag) {
		writefln ("\nBy %s:", boxname);
		print_boxname_flag = 0;
	}
	writefln ("New spammer: %s", address);
}

void scan_mailbox (char[] boxname)
{
	if (! exists (boxname))
		return;

	auto f = new BufferedFile (boxname);
	print_boxname_flag = 1;
	char[] line = f.readLine();
	while (! f.eof () || line.length > 0) {
		char[] nextline;

		if (! f.eof ()) {
			nextline = f.readLine();
			if (nextline.length > 0 && (nextline[0] == ' ' ||
			    nextline[0] == '\t')) {
				line ~= " " ~ strip (nextline);
				continue;
			}
		}
		if (line.length >= 9 && icmp (line[0..9], "Received:") == 0) {
			scan_for_host (line, boxname);
			scan_for_email (line, boxname);
		}
		line = nextline;
	}
	f.close();
	f = new BufferedFile (boxname, FileMode.OutNew);
	f.close();
}

//
// Load old list of host names.
// patterns - list for comparison.
//
void load_black_hosts (char[] filename)
{
	if (! exists (filename))
		return;

	auto f = new BufferedFile (filename);
	while (! f.eof ()) {
		char[] line = f.readLine().strip();
		if (blackhosts.length <= blackhosts_cnt)
			blackhosts.length = blackhosts.length + 100;
		blackhosts [blackhosts_cnt++] = line;

		add_host_pattern (line);
	}
	f.close();
}

//
// Load old lost of spam e-mails.
//
void load_black_emails (char[] filename)
{
	if (! exists (filename))
		return;

	auto f = new BufferedFile (filename);
	while (! f.eof ()) {
		char[] address = f.readLine().strip();
		if (blackemails.length <= blackemails_cnt)
			blackemails.length = blackemails.length + 100;
		blackemails [blackemails_cnt++] = address;
	}
	f.close();
}

char[] reverse (char[] str)
{
	int len = str.length;
	if (len <= 1)
		return str;

	char[] rev;
	rev.length = len;
	rev [len/2] = str [len/2];
	for (int i=0; i<len/2; ++i) {
		rev[i] = str[len-i-1];
		rev[len-i-1] = str[i];
	}
	return rev;
}

char[] common_head (char[] a, char[] b)
{
	char[] head = a.dup;

	for (int n=0; n<head.length; ++n)
		if (a[n] != b[n]) {
			head.length = n;
			break;
		}

	if (head.count(".") < 2)
		return "";
	head.length = head.rfind ('.') + 1;
	return head;
}

void compact_hosts ()
{
	char[][] fixlist;
	int fixlist_cnt = 0;

	if (blackhosts_cnt <= 1)
		return;

	// Reverse host names.
	for (int n=0; n<blackhosts_cnt; ++n)
		blackhosts[n] = reverse (tolower (blackhosts[n]));

	// Compare adjacent strings.
	blackhosts.length = blackhosts_cnt;
	blackhosts.sort;
	for (int n=1; n<blackhosts_cnt; ++n) {
		char[] head = common_head (blackhosts[n-1], blackhosts[n]);
		if (head.length < 8)
			continue;
		char[] rhead = reverse (head);

		// Replace two strings by common part, ending by dot.
		if (fixlist_cnt <= 0 ||
		    cmp (fixlist[fixlist_cnt-1], rhead) != 0) {
			writefln ("%s, %s\n    -> %s", reverse (blackhosts[n-1]),
				reverse (blackhosts[n]), rhead);
			if (fixlist.length <= fixlist_cnt)
				fixlist.length = fixlist_cnt + 100;
			fixlist [fixlist_cnt++] = rhead;
		}
		blackhosts[n-1] = head;
		blackhosts[n] = head;
	}

	// Reverse host names.
	for (int n=0; n<blackhosts_cnt; ++n)
		if (blackhosts[n].length > 0) {
			blackhosts[n] = reverse (blackhosts[n]);
		}
	for (int n=0; n<fixlist_cnt; ++n) {
		if (blackhosts.length <= blackhosts_cnt)
			blackhosts.length = blackhosts_cnt + 100;
		blackhosts [blackhosts_cnt++] = fixlist[n];
	}
	blackhosts.length = blackhosts_cnt;
	blackhosts.sort;
}

char[] common_email (char[] a, char[] b)
{
	char[] head = a.dup;

	for (int n=0; n<head.length; ++n)
		if (a[n] != b[n]) {
			head.length = n;
			break;
		}

	int dog = head.rfind ('@');
	if (dog >= 0)
		head.length = dog;
	else {
		if (head.count(".") < 2)
			return "";
		head.length = head.rfind ('.');
	}
	return head;
}

int white_email (char[] address)
{
	foreach (char[] s; WHITEMAILS)
		if (address.cmp (s) == 0)
			return 1;
	return 0;
}

void compact_emails ()
{
	if (blackemails.length <= 1)
		return;

	// Reverse host names.
	for (int n=0; n<blackemails.length; ++n)
		blackemails[n] = reverse (tolower (blackemails[n]));

	// Compare adjacent strings.
	blackemails.length = blackemails_cnt;
	blackemails.sort;
	for (int n=1; n<blackemails.length; ++n) {
		char[] head = common_email (blackemails[n-1], blackemails[n]);
		if (head.length < 7)
			continue;
		char[] rhead = reverse (head);
		if (white_email (rhead))
			continue;

		// Replace two strings by common part, ending by dot.
		if (n >= 2 && head.cmp (blackemails[n-2]) != 0)
			writefln ("%s\n%s\n    -> %s", reverse (blackemails[n-1]),
				reverse (blackemails[n]), rhead);
		blackemails[n-1] = head;
		blackemails[n] = head;
	}

	// Reverse host names.
	for (int n=0; n<blackemails.length; ++n)
		blackemails[n] = reverse (tolower (blackemails[n]));

	blackemails.sort;
}

void append_logemails (char[] address)
{
	for (int n=0; n<logemails_cnt; ++n)
		if (address.cmp (logemails[n]) == 0)
			return;

	if (logemails.length <= logemails_cnt)
		logemails.length = logemails_cnt + 100;
	logemails [logemails_cnt++] = address;
}

char[] gzreadline (gzFile fd)
{
	char[] buf;
	int buf_cnt = 0;
	for (;;) {
		int c = gzgetc (fd);
		if (c == -1)
			return "";
		if (c == '\r')
			continue;
		if (c == '\n')
			break;
		if (buf.length <= buf_cnt)
			buf.length = buf_cnt + 100;
		buf [buf_cnt++] = c;
	}
	buf.length = buf_cnt;
	return buf;
}

void append_loghosts (char[] address)
{
	if (loghosts.length <= loghosts_cnt)
		loghosts.length = loghosts_cnt + 100;
	loghosts [loghosts_cnt++] = address;
}

void sort_loghosts ()
{
	loghosts.length = loghosts_cnt;
	loghosts.sort;
	int t = 0;
	for (int f=0; f<loghosts.length; f++) {
		if (t == f || loghosts[t].cmp (loghosts[f]) == 0)
			continue;
		loghosts[t++] = loghosts[f];
	}
	loghosts.length = t;
	loghosts_cnt = t;
}

void parse_log_line (char[] line)
{
	if (line.find ("regex") < 0)
		return;

	static RegExp reject_envfrom, reject_connect, tempfail_received;
	if (! reject_envfrom)
		reject_envfrom = new RegExp (".*REJECT ENVFROM <([^>]*)>: Spam.*");
	if (! reject_connect)
		reject_connect = new RegExp (".*: ([^: ]*): REJECT CONNECT.*");
	if (! tempfail_received)
		tempfail_received = new RegExp (".*TEMPFAIL Received from ([^ ]*) .*");

	char[] address = reject_envfrom.replace (line, "$1").strip().tolower();
	if (address.length > 1 && address.find (' ') < 0)
		append_logemails (address);

	address = reject_connect.replace (line, "$1").strip().tolower();
	if (address.length > 1 && address.find (' ') < 0) {
		append_loghosts (address);
	}
	address = tempfail_received.replace (line, "$1").strip().tolower();
	if (address.length > 1 && address.find (' ') < 0) {
		append_loghosts (address);
	}
}

//
// Read log file and add rejected e-mails to logemails.
//
void parse_log (char[] filename)
{
	BufferedFile fd;
	try {
		fd = new BufferedFile (filename);
	} catch {
		return;
	}
	//writefln ("*** Scanning: %s", filename);
	while (! fd.eof()) {
		parse_log_line (fd.readLine());
	}
	fd.close();
}

void parse_gzlog (char[] filename)
{
	gzFile fd;
	try {
		fd = gzopen (toStringz (filename), toStringz ("r"));
	} catch {
		return;
	}
	//writefln ("*** Scanning: %s", filename);
	while (! gzeof (fd)) {
		parse_log_line (gzreadline (fd));
	}
	gzclose (fd);
}

int match_email (char[] pattern, char[] address)
{
	// Remove trailing dot.
	if (address.length > 0 && address [address.length-1] == '.')
		address = address [0 .. address.length-1];

	while (address.cmp (pattern) != 0) {
		// Cut on first dot or @.
		int i = address.find ('@');
		if (i < 0) {
			i = address.find ('.');
			if (i < 0)
				return 0;
		}
		address = address [i+1 .. address.length];
	}
	return 1;
}

int log_contains_email (char[] pattern)
{
	for (int n=0; n<logemails_cnt; ++n)
		if (match_email (pattern, logemails[n]))
			return 1;
	writefln ("Cleaned e-mail: %s", pattern);
	return 0;
}

int log_contains_host (char[] pattern)
{
	int i = pattern.find ('#');
	if (i >= 0)
		pattern = pattern [0 .. i];
	pattern = pattern.strip();
	if (pattern.length <= 0)
		return 1;
	char last = pattern [pattern.length - 1];
	if (last < 'a' || last > 'z')
		last = 0;
	pattern = pattern.replace (".", "\\.");
	pattern = pattern.replace ("?", ".") ~ "$";
	auto re = new RegExp (pattern);
	for (int n=0; n<loghosts_cnt; ++n) {
		if (loghosts[n].length < 1)
			continue;
		if (last && loghosts[n] [loghosts[n].length - 1] != last)
			continue;
		if (re.find (loghosts[n]) >= 0) {
			return 1;
		}
	}
	writefln ("Cleaned host: %s", pattern);
	return 0;
}

void enable_read_access (char[] filename)
{
	char *namez = toStringz (filename);

	std.c.unix.unix.struct_stat st;
	if (std.c.unix.unix.stat (namez, &st) < 0)
		return;

	std.c.unix.unix.chmod (namez,
		cast(std.c.unix.unix.mode_t) (st.st_mode | 0444));
}

//
// Save filtered list of spammer e-mails.
//
void write_black_emails ()
{
	int active_spammers = 0;
	int killed_spammers = 0;

	if (exists (EMAIL_BLACK_LIST ~ "-new"))
		std.file.remove (EMAIL_BLACK_LIST ~ "-new");

	auto f = new BufferedFile (EMAIL_BLACK_LIST ~ "-new", FileMode.Out);
	blackemails.length = blackemails_cnt;
	blackemails.sort;
	int n;
	for (n=0; n<blackemails.length; ++n) {
		if (blackemails[n].length < 1) {
			continue;
		}
		if (n > 0 && blackemails[n].cmp (blackemails[n-1]) == 0) {
			continue;
		}
		if (remove_unused_flag &&
		    ! log_contains_email (blackemails[n])) {
			++killed_spammers;
			continue;
		}
		++active_spammers;
		f.writefln ("%s", blackemails[n]);
	}
	f.close();
	//writefln ("Active e-mails: %d", active_spammers);
	//writefln ("Cleaned e-mails: %d", killed_spammers);

	if (exists (EMAIL_BLACK_LIST))
		std.file.rename (EMAIL_BLACK_LIST, EMAIL_BLACK_LIST ~ "~");

	enable_read_access (EMAIL_BLACK_LIST ~ "-new");
	std.file.rename (EMAIL_BLACK_LIST ~ "-new", EMAIL_BLACK_LIST);
}

//
// Save filtered list of spam hosts.
//
void write_black_hosts ()
{
	int active_spammers = 0;
	int killed_spammers = 0;

	if (exists (HOST_BLACK_LIST ~ "-new"))
		std.file.remove (HOST_BLACK_LIST ~ "-new");

	auto f = new BufferedFile (HOST_BLACK_LIST ~ "-new", FileMode.Out);
	blackhosts.length = blackhosts_cnt;
	blackhosts.sort;
	for (int n=0; n<blackhosts.length; ++n) {
		if (blackhosts[n].length < 1)
			continue;
		if (n > 0 && blackhosts[n].cmp (blackhosts[n-1]) == 0)
			continue;
		if (remove_unused_flag &&
		    ! log_contains_host (blackhosts[n])) {
			++killed_spammers;
			continue;
		}
		++active_spammers;
		f.writeExact (blackhosts[n].ptr, blackhosts[n].length);
		f.write ('\n');
	}
	f.close();
	//writefln ("Active hosts: %d", active_spammers);
	//writefln ("Cleaned hosts: %d", killed_spammers);

	if (exists (HOST_BLACK_LIST))
		std.file.rename (HOST_BLACK_LIST, HOST_BLACK_LIST ~ "~");

	enable_read_access (HOST_BLACK_LIST ~ "-new");
	std.file.rename (HOST_BLACK_LIST ~ "-new", HOST_BLACK_LIST);
}

void touch (char[] filename)
{
	char c;
	auto f = new File (filename, FileMode.In + FileMode.Out);

	try {
		f.read (c);
		f.seek (0L, SeekPos.Set);
		f.write (c);
	} catch {
		// Nothing to do.
	} finally {
		f.close();
	}
}

void main (char[][] args)
{
	foreach (argv; args [1..args.length]) {
		if (argv == "--clean")
			++remove_unused_flag;
	}

	// Read black lists of hosts and e-mails.
	load_black_hosts (HOST_BLACK_LIST);
	load_black_emails (EMAIL_BLACK_LIST);

	// Parse spam mailboxes and add new values to black lists.
	scan_mailbox ("/var/folders/vak/- Spam");
	scan_mailbox ("/var/folders/rik/- Spam");
	scan_mailbox ("/var/folders/ly/- Spam");
	scan_mailbox ("/var/folders/vich/- Spam");
	scan_mailbox ("/var/folders/sak/- Spam");
	scan_mailbox ("/var/folders/leo/- Spam");
	scan_mailbox ("/var/mail/aaa");

	// Find similar host and e-mail values and merge them into patterns.
	compact_hosts ();
	compact_emails ();

	if (remove_unused_flag) {
		// Read logs.
		parse_log ("/var/log/maillog");
		parse_gzlog ("/var/log/maillog.0.gz");
		parse_gzlog ("/var/log/maillog.1.gz");
		parse_gzlog ("/var/log/maillog.2.gz");
		parse_gzlog ("/var/log/maillog.3.gz");
		parse_gzlog ("/var/log/maillog.4.gz");
		parse_gzlog ("/var/log/maillog.5.gz");
		parse_gzlog ("/var/log/maillog.6.gz");
		parse_gzlog ("/var/log/maillog.7.gz");
		parse_gzlog ("/var/log/maillog.8.gz");
		parse_gzlog ("/var/log/maillog.9.gz");
		sort_loghosts ();
		//writefln ("Total %d e-mails, %d hosts", logemails_cnt, loghosts_cnt);
	}
	// Save black lists.
	write_black_emails ();
	write_black_hosts ();

	// Kick sendmail.
	touch ("/etc/mail/milter-regex.conf");
}
